#!/usr/bin/python

# -*- coding: utf-8 -*-
#
# lots of hints and some numerical values taken from
# Virtual Air Guitar (Matti Karjalainen, Teemu Maki-Patola, Aki Kanerva, Antti Huovilainen)
#
# Special edition for OneTriode by NaLex
# V4.0 (new tube model)

import sys
from pylab import *
from scipy.optimize import newton


# Params
mVp = 334
mRp = 100e3
mRi = 68e3

# Input voltage
low = -20
high = 250

# Table size
tsStep = 100 # recommended 200
tabSize = int((abs(low) + high) * tsStep + 1)

# Path to file
filePath = '' # C:\\Ftube\\

# File name prefix
prefix = '12AX7'

# Print step
pStep = 100

# Maximum number of iterations
maxIter = int(900e3)

# 0 - calc_Ftube, 1 - plot_Ftube
mode = 0


#
#                                 o V+         
#                                 |            
#                                 |            
#                                 |            
#                                +-+           
#                                | |           
#                             Rp | |           
#                                | |    Co     
#                                +-+           
#                                 |     ||       
#                                 +-----++---o Vo
#                              Vp |     ||
#                                 |
#     Ci                        --+--  
#                    Ri        /--+--  
#     ||  Vi      +-----+  Vg /         12AX7
# o---++--+-------|     |-----+- - - - 
#     ||  |       +-----+     \        
#         |                    \/----
#         |                    /-----
#         |                   /
#        +-+               Vk |
#        | |                  +------+    
#        | |                  |      |    
#        | |                  |      |    
#        +-+                 +-+     |    
#         |                  | |    -+-   
#         |               Rk | |    -+- Ck
#         |                  | |     |
#       -----                +-+     |
#        ---                  |      |
#         o                   |      |
#                             |      |
#                           -----  -----
#                            ---    ---
#                             o      o
#

def Igk_Vgk(Vgk):
    """gate current as function of gate-kathode voltage"""
    # 12AX7 Tung Sol (Koren-TrueVAL)
    E2 = Vgk+0.78
    # GP 2 3 VALUE={0.000028*(PWR(V(2,3)+0.78,4)+PWRS(V(2,3)+0.78,4))}
    return 0.000028*(abs(E2)**4 + sign(E2)*abs(E2)**4)

def Igk_Vgk2(Vgk):
    """gate current as function of gate-kathode voltage"""
    # 12AX7 (Koren-Rst7)
    E2 = log(1+exp(Vgk))-0.5
    # GP 2 3 VALUE={5E-4*(PWR(V(77,0),1.5)+PWRS(V(77,0),1.5))/2}
    return 5e-4*(abs(E2)**1.5 + sign(E2)*abs(E2)**1.5)/2

def Igk_Vgk1(Vgk):
    """gate current as function of gate-kathode voltage"""
    return exp(7.75*Vgk-10.3)

def Ipk2(Vgk, Vpk):
    """plate current as function of gate-kathode voltage and plate-kathode voltage"""
    # 12AX7 Tung Sol (Koren-TrueVAL)
    mu = 102.0
    kx = 1.5
    kg1 = 1500.0
    kp = 1100.0
    kvb = 300.0
    #E1 = Vpk/kp*log(1+exp(kp*(1/mu+Vgk/sqrt(kvb+Vpk*Vpk))))
    E1 = Vpk/kp*log(1+exp(kp*(1/mu+(0.32+Vgk)/sqrt(kvb+Vpk*Vpk))))
    # G1 1 3 VALUE={(PWR(V(7),EX)+PWRS(V(7),EX))/KG1}
    return (abs(E1)**kx + sign(E1)*abs(E1)**kx)/kg1

def Ipk(Vgk, Vpk):
    """plate current as function of gate-kathode voltage and plate-kathode voltage"""
    mu = 100.0
    kx = 1.4
    kg1 = 1060.0
    kp = 600.0
    kvb = 300.0
    E1 = Vpk/kp*log(1+exp(kp*(1/mu+Vgk/sqrt(kvb+Vpk*Vpk))))
    # G1 1 3 VALUE={(PWR(V(7),EX)+PWRS(V(7),EX))/KG1}
    return (abs(E1)**kx + sign(E1)*abs(E1)**kx)/kg1

def Ipk1(Vgk, Vpk):
    """plate current as function of gate-kathode voltage and plate-kathode voltage"""
    mu = 100.0
    kx = 1.4
    kg1 = 1060.0
    kp = 600.0
    kvb = 300.0
    E1 = Vpk/kp*log(1+exp(kp*(1/mu+Vgk/sqrt(kvb+Vpk*Vpk))))
    return 2*E1**kx/kg1*(E1>0.0)

def Ftube(Vi, Ri, Ipk):
    """calculate output voltage of a tube circuit as function of input voltage
    Vi  array of input voltages
    Ri  value of resistor Ri
    Ipk plate current Ip as function of Vgk and Vpk
    """
    def fi(Vgk, Vi, Ri):
        return Vi - Vgk - Ri * Igk_Vgk(Vgk) # sum of voltages -> zero
    Vgk = newton(fi, 1e-3, args=(Vi, Ri), maxiter=maxIter)   # Vgk(Vi)
    def fp(Vpk, Vgk, Ipk):
        return Vpk + mRp * Ipk(Vgk, Vpk) - mVp # Vp 250V, Rp 100k

    l = 110.0 # start value for iteration
    return newton(fp, l, args=(Vgk, Ipk), maxiter=maxIter)   # Vpk(Vgk)

def calc_ftube(Ri, st, en, cnt):
    """calculate table Vp(Vi)"""
    Vi = linspace(st,en,cnt)
    Vp = zeros_like(Vi)
    for i, v in enumerate(Vi):
        Vp[i] = Ftube(v, Ri, Ipk)
    return Vi, Vp

def calc_ftube2(Ri, st, en, cnt):
    """calculate table Vp(Vi)"""
    Vi = linspace(st,en,cnt)
    Vp = zeros_like(Vi)
    idx = 1
    idxCnt = 0
    fileName = prefix + "  Vp=%dV  Rp=%dK  Ri=%dK" % (mVp, (mRp/1e3), (mRi/1e3))
    fullFilePath = filePath + fileName + '.txt'
    file = open(fullFilePath, 'w')
    print ("Process...")
    for i, v in enumerate(Vi):
        Vp[i] = Ftube(v, Ri, Ipk)
        VpIdx = round(Vp[i],10)
        file.write(str(VpIdx) + '\n')
        if idxCnt >= pStep:
            print (idx, VpIdx)
            idxCnt = 0
        idx += 1
        idxCnt += 1
    file.write(str(mVp) + '\n')
    file.write(str(mRp) + '\n')
    file.write(str(mRi) + '\n')
    file.write(str(low) + '\n')
    file.write(str(high) + '\n')
    file.write(str(tabSize))
    file.close()
    print ()
    print ("File: " + fullFilePath)
    return Vi, Vp

def write_ftube_table(Ri, Vi, Vp):
    """write C source"""
    sys.stdout.write("\t{ // Ri = %dk\n" % (Ri/1e3))
    sys.stdout.write('\t%g,%g,%g, {' % (Vi[0], Vi[-1], (len(Vi)-1)/(Vi[-1]-Vi[0])))
    s = ""
    for i, v in enumerate(Vi):
        if i % 5 == 0:
            sys.stdout.write(s+"\n\t")
            s = ""
        sys.stdout.write(s+str(Vp[i]))
        s = ","
    sys.stdout.write("\n\t}}")

def write_tables():
    sys.stdout.write("struct table1d tubetable[] = {\n")
    s = ""
    for Ri in 68e3, 250e3:
        sys.stdout.write(s)
        s = ",\n"
        Vi, Vp = calc_ftube(Ri, -5, 5, 2001)
        write_ftube_table(Ri, Vi, Vp)
    sys.stdout.write("\n};\n")

def check_table_accuracy(Ri, st, en, cnt):
    """maximal table error at half interval"""
    Vi, Vp = calc_ftube(Ri, st, en, cnt)
    d = zeros((len(Vp)-1,))
    for i in range(1,len(Vp)):
        vi = (Vi[i-1]+Vi[i])/2
        v = Ftube(vi, Ri, Ipk)
        d[i-1] = (0.5*(Vp[i-1]+Vp[i]) - v)/v
    return max(d)

def display_accuracy():
    for Ri in 250e3, 68e3:
        print ("Ri=%dk: %g" % (Ri/1e3, check_table_accuracy(Ri, -5, 5, 2001)))

def plot_Ftube():
    Vi = linspace(-5,5)
    Vp = zeros_like(Vi)
    for Ri in 250e3, 68e3:
        for i, v in enumerate(Vi):
            Vp[i] = Ftube(v, Ri, Ipk)
        plot(Vi, Vp, label="Ri=%dk" % (Ri/1e3))
    xlabel("Vik")
    ylabel("Vp")
    legend()
    axis
    show()
    
def plot_Ftube2():
    Vi = linspace(low,high)
    Vp = zeros_like(Vi)
    for i, v in enumerate(Vi):
        Vp[i] = Ftube(v, mRi, Ipk)
    plot(Vi, Vp, label="Ri=%dk" % (mRi/1e3))
    xlabel("Vik")
    ylabel("Vp")
    legend()
    axis
    show()

def main():
    if len(sys.argv) == 2:
        arg = sys.argv[1]
    else:
        arg = ""
    if arg == "plot":
        plot_Ftube()
    elif arg == "accuracy":
        display_accuracy()
    elif arg == "table":
        write_tables()
    else:
        print ("usage: %s plot|accuracy|table" % sys.argv[0])

def main2():
    if mode == 0:
        tsLimit = 100e3 # table size limit
        if (tabSize + 6) > tsLimit:
            print ("Error: Table size > %d" % tsLimit)
        else:
            calc_ftube2(mRi, low, high, tabSize)
    else:
        plot_Ftube2()

if __name__ == "__main__":
    main2()
